{-encapsula la definicin de un objeto yahu customizado: un panel de minas; para separar
la funcionalidad especfica del resto de la aplicacin (juego del buscaminas)-}
module MinePanel (
	MinePanel

	, pressed	-- :: Event MinePanel (Point -> IO ())
	, flagged	-- :: Event MinePanel (Bool -> Point -> IO ())
	, exploded	-- :: Event MinePanel (Point -> IO ())
	, startGame	-- :: Event MinePanel (Point -> IO ())
	, winGame	-- :: Event MinePanel (IO ())
	, numMines	-- :: Attr MinePanel Int
	, nakedCells	-- :: MinePanel -> IO Int
	, numFlags	-- :: MinePanel -> IO Int
	, numMarks	--  :: MinePanel -> IO Int
	, mpState	--  :: MinePanel -> IO MPState
	, cellDim	-- :: MinePanel -> IO Size
	, numCells	-- :: MinePanel -> IO Int
	
	, minePanel	-- :: (TopLevelWindow w) => Size -> Size -> w -> IO MinePanel
	, resetMP	-- :: MinePanel -> IO ()
	
	, MPState(..)
	) where

import Graphics.UI.Yahu.All
import System.Random
import Data.Array
import Data.List (nub)
import Monad (filterM, foldM, when)
import System.IO

-- | estado global de un panel de minas
data MPState = Reset | Playing | Lose | Win deriving (Eq, Show)

-- | estado de una celda
data CellState = Normal | Naked | Flagged | Queried deriving Eq

-- | Panel de minas
data MinePanel = MinePanel {
		mptPanel :: Panel
		, mptCellSize :: Size
		, mptSize :: Size
		, mptNumCells :: Int
--para simular las celdas usamos pares button/label. Una celda est cubierta cuando es visible
--el botn, y descubierta cuando es visible la etiqueta. Una celda tiene una mina cuando
--su etiqueta tiene el texto "*".
		, mptCells :: Array (Int, Int) (Button, Label)
		, mptState :: State MPState
		, mptOnPressed :: State (Point -> IO ())
--este callback puede ocurrir tanto al ponerse un flag (True) como al eliminarse (False)
		, mptOnFlagged :: State (Bool -> Point -> IO ())
		, mptOnExploded :: State (Point -> IO ())
		, mptOnStart :: State (Point -> IO ())
		, mptOnWin :: State (IO ())
		, mptNumMines :: State Int
		, mptNumNakeds :: State Int
		, mptNumFlags :: State Int
		, mptNumMarks :: State Int
		}

-- | helper State -> Atributo
auxAttr :: (p -> State a) -> Attr p a
auxAttr toST = newAttr (\p-> toST p?:value, \p new-> toST p!:value =: new)

-- | delegacin de un atributo del nuevo objeto en uno de sus componentes
delegateAttr :: (w' -> w) -> Attr w a -> Attr w' a
delegateAttr trans attr = newAttr (\p->trans p?:attr, \p new->trans p!:attr =: new)

-- | helper State -> Evento
auxEvent :: (p -> State a) -> Event p a
auxEvent toST = newEvent toST ((?:value), \st new-> st!:value =: new)

-- | helper State -> IO
auxVal :: (p -> State a) -> (p -> IO a)
auxVal toST p = toST p?:value

-----------------------------------------------------------------------------
--interfaz con el exterior

--instancia delegada necesaria para la de Dimensions
instance Widget MinePanel where
  toWXWin = toWXWin.mptPanel

--instancia delegada
instance Dimensions MinePanel where
  dimX = delegateAttr mptPanel dimX
  dimY = delegateAttr mptPanel dimY
  dim = delegateAttr mptPanel dim

-- instancia "casi" delegada
instance Finalizable MinePanel where
  --nos morimos lejos de toda mirada, para que no se vea como cae muerta fila tras fila de
  --botones
  destroy (MinePanel{mptPanel=p}) = (p!:shown =: False) >> destroy p
  varDestroyed (MinePanel{mptPanel=p}) = varDestroyed p

instance EventListener MinePanel

--instancia delegada
instance Able MinePanel where
  enabled = delegateAttr mptPanel enabled

--instancia delegada
instance Contained MinePanel where
  toLayout = toLayout.mptPanel

-- | evento de pinchar en una celda
pressed :: Event MinePanel (Point -> IO ())
pressed = auxEvent mptOnPressed

-- | evento de colocar una marca de mina a una celda
flagged :: Event MinePanel (Bool -> Point -> IO ())
flagged = auxEvent mptOnFlagged

-- | evento de perder el juego
exploded :: Event MinePanel (Point -> IO ())
exploded = auxEvent mptOnExploded

-- | evento de empezar el juego
startGame :: Event MinePanel (Point -> IO ())
startGame = auxEvent mptOnStart

-- | evento de ganar el juego
winGame :: Event MinePanel (IO ())
winGame = auxEvent mptOnWin

-- | n de minas
numMines :: Attr MinePanel Int
numMines = auxAttr mptNumMines

-- | nmero de celdas descubiertas
nakedCells :: MinePanel -> IO Int
nakedCells = auxVal mptNumNakeds

-- | nmero de celdas marcadas con flags
numFlags :: MinePanel -> IO Int
numFlags = auxVal mptNumFlags

-- | nmero de celdaas marcadas con interrogacin
numMarks  :: MinePanel -> IO Int
numMarks = auxVal mptNumMarks

-- | estado global
mpState  :: MinePanel -> IO MPState
mpState = auxVal mptState

-- | tamao en nmero de columnas/filas de celdas
cellDim :: MinePanel -> Size
cellDim = mptSize

-- | nmero de celdas
numCells :: MinePanel -> Int
numCells = mptNumCells

{--------------------------------------------------------------------------------------}

-- | fold sobre rangos de arrays
foldArr :: Ix a => (a -> b -> c -> c) -> c -> (a, a) -> Array a b -> c
foldArr f init limits array = foldr (\pos xs-> f pos (array!pos) xs) init (range limits)

-- | sequence_ sobre rangos de arrays
sequenceArr_ :: Ix a => (a -> b -> IO ()) -> (a,a) -> Array a b -> IO ()
sequenceArr_ f = foldArr (\pos val acc-> acc >> f pos val) (return ())

-- | variamos estrictamente el valor de una variable, para que no se acumule luego el trabajo
changeStrict :: State a -> (a -> a) -> IO ()
changeStrict var fun = var!:value <<: (\val-> let next = fun val in next `seq` return next)

-- | Al asignar el texto se recalcula el tamao ptimo, aunque no usemos layouts. Como
--queremos mantenerlo siempre igual, usamos este apao
setText :: (Text w, Dimensions w) => String -> Size -> w -> IO ()
setText str cellsize obj = do {
	; obj!:text =: str
	; obj!:dim =: cellsize
	}

-- | lista de celdas vecinas de otra celda
cellsvecinas :: (Int,Int) -> (Int,Int) -> [(Int,Int)]
cellsvecinas (ox,oy) (cx,cy) = [ (x,y) | x <- [minx..maxx], y <- [miny..maxy], x/=ox || y/=oy ]
 where
 	oxm1 = ox+1
 	oym1 = oy+1
 	minx = if ox == 0 then ox else ox-1
 	maxx = if oxm1 == cx then ox else oxm1
 	miny = if oy == 0 then oy else oy-1
 	maxy = if oym1 == cy then oy else oym1

-- | cdigo de colores
num2Col :: Int -> Color
num2Col 0 = grey
num2Col 1 = cyan
num2Col 2 = green
num2Col 3 = red
num2Col 4 = blue
num2Col 5 = darkgrey
num2Col 6 = yellow
num2Col 7 = magenta
num2Col 8 = black

-- | tomado de pepe gallardo con modificaciones. Genera una lista no repetida
--de posiciones aleatorias, con el aadido de que no pone minas sobre una de las celdas
rnds :: (Int,Int) -> (Int,Int) -> IO [(Int,Int)] 
rnds (mx,my) forbidden = do {
	; g <- newStdGen
	; let xs = randomRs (0,mx-1) g
	; g' <- newStdGen
	; let ys = randomRs (0,my-1) g'
	--aadimos la posicin prohibida a la cabeza, y luego la descartamos. Es donde hemos
	--pinchado (no queremos pinchar un mina justo cuando empezamos)
	; let sinrepets = nub (forbidden:(zip xs ys))
	; return (tail sinrepets)
	}

-- | helper para generar las minas
genMineField :: (a -> IO ()) -> (Int, Int) -> Int -> (Int, Int) -> Array (Int, Int) a -> IO ()
genMineField notify_acc sizefield numminas forbiddenpos array = do {
	; minasparasiempre <- rnds sizefield forbiddenpos
	; let minas = take numminas minasparasiempre
	; mapM_ (\pos-> notify_acc (array!pos)) minas
	}

-- | realiza una accin entre bastidores, sin que se vea cmo cambia el panel
withPanelOff :: Panel -> IO () -> IO ()
withPanelOff panel acc = do {
	; panel!:shown =: False
	; acc
	; panel!:shown =: True
	}

-- | resetea un campo de minas, dejndolo listo para jugar
resetMP :: MinePanel -> IO ()
resetMP (MinePanel{mptPanel=panel, mptCellSize=cellsize, mptSize=(Size cols rows)
 , mptCells=cells, mptState=state, mptNumNakeds=nakeds, mptNumFlags=flags
 , mptNumMarks=marks}) = withPanelOff panel (do {
	; let rangeAll = ((0,0), (cols-1,rows-1))
	; sequenceArr_ (\ pos@(x,y) (but, lab) -> do {
		; setText "" cellsize but
		; setText "" cellsize lab
		; but!:enabled =: True
		; lab!:shown =: False
		; lab!:bgColor =: lightgrey
		; lab!:fgColor =: black
		; but!:shown =: True
		})
		rangeAll cells
	; nakeds!:value =: 0
	; flags!:value =: 0
	; marks!:value =: 0
	; state!:value =: Reset
	})

-- | pulsa una celda, con los posibles resultados: termina el juego (bien o mal) o la descubre
--(junto a sus vecinas si no tiene minas vecinas)
pressCellMP :: Point -> MinePanel -> IO ()
pressCellMP p@(Point x y) (MinePanel{mptPanel=panel, mptCellSize=cellsize
 , mptNumCells=numcells, mptCells=cells, mptSize=(Size cx cy), mptOnPressed=onpressed
 , mptState=globalst, mptOnExploded=onexploded, mptOnWin=onwin, mptNumMines=minas
 , mptNumNakeds=nakeds}) = do {
	; nopressed <- but?:shown
	; when nopressed (do {
		; txt <- lab?:text
		; case txt of
			"" -> do { -- sin minas
				; nakeCell pos
				; num <- nakeds?:value
				; numM <-minas?:value
				; if num+numM==numcells then do {
					; globalst!:value =: Win
					; callback <- onwin?:value
					; callback
				} else do {
					; callback <- onpressed?:value
					; callback p
				}}
			"*" -> do { -- con minas, hemos perdido
				; lab!:bgColor =: red
				; nakeMines
				; globalst!:value =: Lose
				; callback <- onexploded?:value
				; callback p
				}
			_ -> return () --puede darse el caso de entrar por aqu; no hacemos nada
		})
	}
 where
 	pos = (x,y)
 	(but,lab) = cells!pos
 	
 	--desnuda una celda (y quizs sus vecinas)
 	nakeCell :: (Int,Int) -> IO ()
 	nakeCell pos@(x, y) = do {
 		; let (but,lab) = cells!pos
 		; let vecinas = cellsvecinas pos (cx,cy)
 		; num <- minasVecinas vecinas
 		--calculamos el nmero de vecinas con minas in situ
 		; setText (if num==0 then "" else show num) cellsize lab
 		; lab!:fgColor =: num2Col num
 		; lab!:shown =: True
 		; but!:shown =: False
 		; changeStrict nakeds (+1)
 		; when (num==0) (descubreVecinas vecinas)
 		}
 	
 	-- | devuelve el nmero de minas vecinas
 	minasVecinas :: [(Int,Int)] -> IO Int
 	minasVecinas = foldM (\num pos-> do {
 		; let (_,lab) = cells!pos
 		; txt <- lab?:text
 		; return (num+if txt=="*" then 1 else 0)
 		}) 0
 	
 	-- | descubre las celdas vecinas no descubiertas todava
 	descubreVecinas :: [(Int,Int)] -> IO ()
 	descubreVecinas = mapM_ (\pos -> do {
 		; let (but,_)= cells!pos
 		; notnaked <- but?:shown
 		; when notnaked (nakeCell pos)
 		})

 	-- | descubre las minas
 	nakeMines :: IO ()
 	nakeMines = sequenceArr_ (\_ (but,lab)-> do {
 		; txtb <- but?:text
 		; txtl <- lab?:text
 		; case txtb of
 			"M" -> when (txtl/="*") (setText "X" cellsize but)
 			_ -> when (txtl=="*") (do {
 				; but!:shown =: False
 				; lab!:shown =: True
 				})
 		})
 		((0,0), (cx-1,cy-1)) cells

whenJust :: Maybe a -> (a -> IO ()) -> IO ()
whenJust mb acc = case mb of
	Nothing -> return ()
	Just a -> acc a

-- | genera una panel de minas (es un chorizo muy grande porque no delegamos en auxiliares
--la configuracin de los eventos de botones y etiquetas). Toma tambin una barra de
--progreso en la que mostrar su progreso porque puede ser un proceso lento
minePanel :: (From0Range g, TopLevelWindow w) => Size -> Size -> Maybe g -> w -> IO MinePanel
minePanel cellsize@(Size sx sy) numceldas@(Size cx cy) mbbarraprogreso parent = do {
	; let cxXcy = cx*cy
	; let ifBarraProgreso = whenJust mbbarraprogreso
	; ifBarraProgreso (\barraprogreso-> do {
		; barraprogreso!:currentPos =: 0
		; barraprogreso!:maxRange =: cxXcy*2 --dos pasadas: creacin e inicializacin
		})
	--creamos todo
	; pan <- panel [] parent
	; onpressed <- state [value =: \_->return ()]
	; onflagged <- state [value =: \_ _->return ()]
	; onexploded <- state [value =: \_->return ()]
	; onstart <- state [value =: \_->return ()]
	; onwin <- state [value =: return ()]
	; globalst <- state [value =: Reset]
	; minas <- state [value =: 1]
	; nakeds <- state [value =: 0]
	; flags <- state [value =: 0]
	; marks <- state [value =: 0]
	; let mkcell (px,py) = do {
		; let posicion = point (px*sx) (py*sy)
		; but <- button [pos =: posicion, dim =: cellsize, fontWeight =: FontWeightBold] pan
		; lab <- label' [justify =:. CenterJust, border =:. SimpleB]
						[pos =: posicion, dim =: cellsize, shown =: False, fontWeight =: FontWeightBold] pan
		; ifBarraProgreso (\barraprogreso-> barraprogreso!:currentPos <: (+1))
		; return (but,lab)
		}
	; lcells <- sequence [ mkcell (x,y) | x <- [0..(cx-1)], y <- [0..(cy-1)] ]
	; let rangeAll = ((0,0),(cx-1,cy-1))
	; let cells = listArray rangeAll lcells
	; let ret = MinePanel pan cellsize numceldas cxXcy cells globalst onpressed onflagged onexploded onstart onwin minas nakeds flags marks
	--preparamos los manejadores de eventos
	; sequenceArr_ (\pos@(x,y) (but,lab)-> do {
		; let posicion = point x y
		--doble click derecho sobre celda descubierta: si se han colocado tantos flags
		--como minas tiene alrededor la celda, se descubren las restantes (para bien o para mal)
		; lab!:on (right $ doubleClick mouse) =: \_-> do {
			; st <- globalst?:value
			; when (st==Playing) (do{
				; txt <- lab?:text
				--intentamos parsear un nmero (que nosotros habremos asignado
				--previamente en el auxiliar de pressCell, nakeCell) del texto
				--de la etiqueta
				; case reads txt of
					[] -> return ()
					[(num,_)] -> do {
						; let vecinas = cellsvecinas pos (cx,cy)
						; let minasVecinas = foldM (\num pos-> do {
							; let (but,_) = cells!pos
							; txt <- but?:text
							; return (num+if txt=="M" then 1 else 0)
							}) 0
						; num' <- minasVecinas vecinas
						; when (num==num') (mapM_ (\pos@(x,y)-> do {
							; let (but,_) = cells!pos
							; txt <- but?:text
							; st <- globalst?:value
							; when ((txt/="M")&&(st/=Lose)) (pressCellMP (point x y) ret)
							}) vecinas)
						}
				})
			; return True
			}
		--slo permitimos que se ejecute el evento action si el botn no est marcado con un flag
		; but!:on (left mouse) =: \_-> do {
			; txt <- but?:text
			; return (txt/="M")
			}
		--slo permitimos que se ejecute el evento action si el botn no est marcado con un flag
		; but!:on (left $ doubleClick mouse) =: \_-> do {
			; txt <- but?:text
			; return (txt/="M")
			}
		--presionamos el botn; descubrimos la celda (y si la partida no estaba empezada,
		--la empezamos, teniendo cuidado de no poner una mina en la celda pulsada)
		; but!:command =: do {
			; st <- globalst?:value
			; case st of
				Playing -> pressCellMP posicion ret
				Reset -> do {
					; numminas <- minas?:value
					; genMineField (\(_,lab)-> setText "*" cellsize lab) (cx,cy) numminas pos cells
					; globalst!:value =: Playing
					; callback <- onstart?:value
					; callback posicion
					; pressCellMP posicion ret
					}
				_ -> return ()
			}
		-- cambiamos el estado de la celda no descubierta: normal, con flag, o con marca
		; but!:on (right mouse) =: \_-> do {
			; st <- globalst?:value
			; let acc = do {
				; txt <- but?:text
				; case txt of
					"" -> do {
						; setText "M" cellsize but
						; changeStrict flags (+1)
						; callback <- onflagged?:value
						; callback True posicion
						}
					"M" -> do {
						; setText "?" cellsize but
						; changeStrict flags (\x->x-1)
						; changeStrict marks (+1)
						; callback <- onflagged?:value
						; callback False posicion
						}
					"?" -> do {
						; setText "" cellsize but
						; changeStrict marks (\x->x-1)
						}
				}
			; case st of
				Reset -> acc
				Playing -> acc
				_ -> return ()
			; return True
			}
		; ifBarraProgreso (\barraprogreso-> barraprogreso!:currentPos <: (+1))
		})
		rangeAll cells
	--devolvemos el objeto recin sacado del horno (hace falta resetearlo, por cuestiones
	--de tamaos de celdas)
	; return ret
	}


